from pathlib import Path
from time import time

import numpy as np
import pyqtgraph.opengl as gl
from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QSlider,
    QVBoxLayout,
    QWidget,
)

import caliscope.logger
from caliscope.cameras.camera_array import CameraArray
from caliscope.gui.vizualize.camera_mesh import CameraMesh, mesh_from_camera
from caliscope.motion_trial import MotionTrial

logger = caliscope.logger.get(__name__)
# as part of development process I'm just going to import the skeleton in here


class PlaybackTriangulationWidget(QWidget):
    def __init__(self, camera_array: CameraArray, xyz_history_path: Path = None):
        super(PlaybackTriangulationWidget, self).__init__()

        self.camera_array = camera_array
        self.visualizer = TriangulationVisualizer(self.camera_array)
        self.slider = QSlider(Qt.Orientation.Horizontal)

        self.setMinimumSize(500, 500)

        self.place_widgets()
        self.connect_widgets()
        if xyz_history_path is not None:
            self.update_motion_trial(xyz_history_path)
        else:
            self.motion_trial = None

    def place_widgets(self):
        self.setLayout(QVBoxLayout())
        self.layout().addWidget(self.visualizer.scene)
        self.layout().addWidget(self.slider)

    def connect_widgets(self):
        self.slider.valueChanged.connect(self.visualizer.display_points)
        self.slider.valueChanged.connect(self.visualizer.update_segment_lines)

    def update_motion_trial(self, xyz_history_path):
        # self.xyz_history = pd.read_csv(xyz_history)
        tic = time()
        logger.info(f"Beginning to load in motion trial: {time()}")
        self.motion_trial = MotionTrial(xyz_history_path)
        logger.info(f"Motion trial loading complete: {time()} ")
        toc = time()
        logger.info(f"Elapsed time to load: {toc-tic}")

        self.visualizer.update_motion_trial(self.motion_trial)

        if self.motion_trial.is_empty:
            self.slider.setMinimum(0)
            self.slider.setMaximum(100)
        else:
            self.slider.setMinimum(self.motion_trial.start_index)
            self.slider.setMaximum(self.motion_trial.end_index)

    def update_camera_array(self, camera_array: CameraArray):
        self.visualizer.update_camera_array(camera_array)


class TriangulationVisualizer:
    def __init__(self, camera_array: CameraArray):
        self.camera_array = camera_array
        self.build_scene()

    def build_scene(self):
        # constuct a scene if not yet there
        if hasattr(self, "scene"):
            logger.info("Clearing scene in capture volume visualizer")
            self.scene.clear()
        else:
            logger.info("Creating initial scene in capture volume visualizer")
            self.scene = gl.GLViewWidget()

            # the scene camera, not a real Camera
            self.scene.setCameraPosition(distance=4)
        axis = gl.GLAxisItem()
        self.scene.addItem(axis)

        if self.camera_array.all_extrinsics_calibrated():
            self.meshes = {}
            for port, cam in self.camera_array.cameras.items():
                mesh: CameraMesh = mesh_from_camera(cam)
                self.meshes[port] = mesh
                self.scene.addItem(mesh)

        self.scatter = gl.GLScatterPlotItem(
            pos=np.array([0, 0, 0]),
            color=[1, 1, 1, 1],
            size=0.01,
            pxMode=False,
        )

        # will hold a list
        self.segments = {}

        self.scene.addItem(self.scatter)
        self.scatter.setData(pos=None)

    def update_camera_array(self, camera_array: CameraArray):
        self.camera_array = camera_array
        self.build_scene()

    def update_motion_trial(self, motion_trial: MotionTrial):
        logger.info("Updating xyz history in playback widget")
        self.motion_trial: MotionTrial = motion_trial

        if hasattr(self.motion_trial.tracker, "wireframe"):
            for segment_line in self.motion_trial.tracker.wireframe.line_plots.values():
                self.scene.addItem(segment_line)

        self.sync_index = self.motion_trial.start_index
        self.display_points(self.sync_index)

    def display_points(self, sync_index: int):
        """
        sync_index is provided from the dialog and linked to the slider
        it is initially set to the minimum viable sync index
        """

        if self.motion_trial.is_empty:
            self.scatter.setData(pos=None)
        else:
            self.sync_index = sync_index
            logger.debug(f"Displaying xyz points for sync index {sync_index}")
            xyz_coords = self.motion_trial.get_xyz(self.sync_index).point_xyz
            self.scatter.setData(pos=xyz_coords)

    def update_segment_lines(self, sync_index: int):
        self.motion_trial.update_wireframe(sync_index)



import sys
from pathlib import Path

from PySide6.QtCore import Qt, Signal, Slot
from PySide6.QtGui import QIcon, QPainter, QPixmap
from PySide6.QtSvg import QSvgRenderer
from PySide6.QtWidgets import (
    QApplication,
    QCheckBox,
    QDoubleSpinBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSlider,
    QSpinBox,
    QStyle,
    QVBoxLayout,
    QWidget,
)

import caliscope.logger
from caliscope import __root__
from caliscope.controller import Controller
from caliscope.gui.camera_management.camera_display_widget import (
    CameraDataDisplayWidget,
)

logger = caliscope.logger.get(__name__)


def svg_to_pixmap(svg_path: Path, size):
    # Load SVG file
    renderer = QSvgRenderer(str(svg_path))

    # Create an empty QPixmap and fill it with transparent color
    pixmap = QPixmap(size)
    pixmap.fill(Qt.transparent)

    # Render SVG onto QPixmap
    painter = QPainter(pixmap)
    renderer.render(painter)
    painter.end()

    return pixmap


class CustomSlider(QSlider):
    arrowKeyPressed = Signal(int)  # Custom signal

    def __init__(self):
        super(CustomSlider, self).__init__(Qt.Horizontal)
        self._isDragging = False
        self.valueChanged.connect(self.checkForArrowKey)

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            self._isDragging = True
        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        if event.button() == Qt.LeftButton:
            self._isDragging = False
        super().mouseReleaseEvent(event)

    def isDragging(self):
        return self._isDragging

    def isUsingArrowKeys(self):
        return self.hasFocus() and not self._isDragging

    @Slot(int)
    def checkForArrowKey(self, value):
        if self.isUsingArrowKeys():
            self.arrowKeyPressed.emit(value)  # Emit the custom signal


# icons from https://iconoir.com
CAM_ROTATE_RIGHT_PATH = Path(__root__, "caliscope", "gui", "icons", "rotate-camera-right.svg")
CAM_ROTATE_LEFT_PATH = Path(__root__, "caliscope", "gui", "icons", "rotate-camera-left.svg")


class IntrinsicCalibrationWidget(QWidget):
    def __init__(self, controller: Controller, port: int, parent=None):
        super().__init__(parent)
        self.controller = controller
        self.port = port

        self.total_frames = self.controller.get_intrinsic_stream_frame_count(self.port)
        self.frame_image = QLabel(self)
        self.frame_index_label = QLabel(self)
        self.play_button = QPushButton("", self)
        self.play_icon = self.style().standardIcon(QStyle.SP_MediaPlay)
        self.pause_icon = self.style().standardIcon(QStyle.SP_MediaPause)

        self.play_button.setIcon(self.play_icon)
        self.slider = CustomSlider()
        self.slider.setMaximum(self.total_frames - 1)

        self.add_grid_btn = QPushButton("Add Grid")
        self.calibrate_btn = QPushButton("Calibrate")
        self.camera_data_display = CameraDataDisplayWidget(self.port, self.controller)
        self.clear_calibration_data_btn = QPushButton("Clear Data")
        self.toggle_distortion = QCheckBox("Apply Distortion")

        self.cw_rotation_btn = QPushButton(QIcon(str(CAM_ROTATE_RIGHT_PATH)), "")
        self.cw_rotation_btn.setMaximumSize(35, 35)
        self.ccw_rotation_btn = QPushButton(QIcon(str(CAM_ROTATE_LEFT_PATH)), "")
        self.ccw_rotation_btn.setMaximumSize(35, 35)

        self.autocalibrate_btn = QPushButton("Autocalibrate")
        self.target_grid_count_spin = QSpinBox(self)
        self.target_grid_count_spin.setMaximumWidth(40)
        self.target_grid_count_spin.setRange(0, 100)
        self.target_grid_count_spin.setValue(20)
        self.target_grid_count_spin.setSingleStep(1)

        self.board_threshold_spin = QDoubleSpinBox(self)
        self.board_threshold_spin.setMaximumWidth(50)
        self.board_threshold_spin.setRange(0, 1)
        self.board_threshold_spin.setValue(0.5)
        self.board_threshold_spin.setSingleStep(0.1)

        # Create the spinbox
        self.scaling_spin = QSpinBox(self)
        self.scaling_spin.setRange(50, 150)
        self.scaling_spin.setValue(100)
        self.scaling_spin.setSingleStep(5)

        self.is_playing = False

        self.place_widgets()
        self.connect_widgets()

    def place_widgets(self):
        self.layout = QHBoxLayout()
        self.setLayout(self.layout)
        self.layout.addWidget(self.camera_data_display)
        self.right_panel = QVBoxLayout()
        self.right_panel.addWidget(self.frame_image, alignment=Qt.AlignmentFlag.AlignCenter)

        self.rotate_span = QHBoxLayout()
        self.rotate_span.addWidget(self.cw_rotation_btn)
        self.rotate_span.addWidget(self.ccw_rotation_btn)
        self.right_panel.addLayout(self.rotate_span)

        self.auto_control_span = QHBoxLayout()
        self.auto_control_span.addWidget(QLabel("Target Grid Count:"), alignment=Qt.AlignmentFlag.AlignRight)
        self.auto_control_span.addWidget(self.target_grid_count_spin, alignment=Qt.AlignmentFlag.AlignLeft)
        self.auto_control_span.addWidget(QLabel("Board Threshold:"), alignment=Qt.AlignmentFlag.AlignRight)
        self.auto_control_span.addWidget(self.board_threshold_spin, alignment=Qt.AlignmentFlag.AlignLeft)
        self.auto_control_span.addWidget(self.autocalibrate_btn)
        self.right_panel.addLayout(self.auto_control_span)

        self.play_span = QHBoxLayout()
        self.play_span.addWidget(self.play_button)
        self.play_button.setMaximumWidth(35)
        self.play_span.addWidget(self.slider)
        self.right_panel.addLayout(self.play_span)

        self.manual_control_span = QHBoxLayout()
        self.manual_control_span.addWidget(self.add_grid_btn)
        self.manual_control_span.addWidget(self.calibrate_btn)
        self.manual_control_span.addWidget(self.clear_calibration_data_btn)
        self.right_panel.addLayout(self.manual_control_span)

        self.distortion_control_span = QHBoxLayout()
        self.distortion_control_span.addWidget(self.toggle_distortion, alignment=Qt.AlignmentFlag.AlignCenter)
        self.distortion_control_span.addWidget(QLabel("Zoom:"), alignment=Qt.AlignmentFlag.AlignRight)
        self.distortion_control_span.addWidget(self.scaling_spin, alignment=Qt.AlignmentFlag.AlignLeft)
        self.distortion_control_span.addWidget(self.frame_index_label)
        self.right_panel.addLayout(self.distortion_control_span)
        self.layout.addLayout(self.right_panel)

    def connect_widgets(self):
        self.play_button.clicked.connect(self.play_video)
        self.slider.sliderMoved.connect(self.slider_moved)
        self.slider.arrowKeyPressed.connect(self.slider_moved)
        self.add_grid_btn.clicked.connect(self.add_grid)
        self.calibrate_btn.clicked.connect(self.calibrate)
        self.clear_calibration_data_btn.clicked.connect(self.clear_calibration_data)
        self.toggle_distortion.stateChanged.connect(self.toggle_distortion_changed)
        self.ccw_rotation_btn.clicked.connect(self.rotate_ccw)
        self.cw_rotation_btn.clicked.connect(self.rotate_cw)
        self.autocalibrate_btn.clicked.connect(self.autocalibrate)

        self.scaling_spin.valueChanged.connect(self.on_scale_change)
        self.controller.intrinsic_stream_manager.frame_emitters[self.port].ImageBroadcast.connect(self.update_image)
        self.controller.intrinsic_stream_manager.frame_emitters[self.port].FrameIndexBroadcast.connect(
            self.update_index
        )
        self.controller.enable_inputs.connect(self.update_enable_all_inputs)

        # initialize stream to push first frame to widget then hold
        # must be done after signals and slots connected for effect to take hold
        self.controller.play_intrinsic_stream(self.port)

        # self.play_started = True
        self.controller.pause_intrinsic_stream(self.port)
        self.controller.stream_jump_to(self.port, 0)

    def play_video(self):
        # if self.play_started:
        if self.is_playing:
            self.is_playing = False
            self.controller.pause_intrinsic_stream(self.port)
            self.play_button.setIcon(self.play_icon)
        else:
            self.is_playing = True
            self.controller.unpause_intrinsic_stream(self.port)
            self.play_button.setIcon(self.pause_icon)

    def slider_moved(self, position):
        self.controller.stream_jump_to(self.port, position)
        if position == self.total_frames - 1:
            self.controller.pause_intrinsic_stream(self.port)
            self.is_playing = False
            self.play_button.setEnabled(False)
            self.play_button.setIcon(self.play_icon)
        else:
            if not self.play_button.isEnabled():
                self.play_button.setEnabled(True)

    def update_index(self, port, position):
        """
        only update slider with the position when the stream is making it happen
        track user interact with the widget to assess whether user is currently interacting
        with the slider, at which point don't try to programmatically change the position
        """
        if port == self.port:
            self.index = position
            self.frame_index_label.setText(f"Frame Index: {self.index}")
            if self.slider.isDragging() or self.slider.isUsingArrowKeys():
                pass  # don't change slider position as this would create a feedback loop
            else:
                self.slider.setValue(position)
                if position == self.total_frames - 1:
                    self.controller.pause_intrinsic_stream(self.port)
                    self.is_playing = False
                    self.play_button.setEnabled(False)
                    # now paused so only option is play
                    self.play_button.setIcon(self.play_icon)

    def update_image(self, port, pixmap):
        logger.debug(f"Running `update_image` in playback widget for port {self.port}")
        if port == self.port:
            self.frame_image.setPixmap(pixmap)

    def add_grid(self):
        self.controller.add_calibration_grid(self.port, self.index)
        self.controller.stream_jump_to(self.port, self.index)

    def on_scale_change(self, value):
        """
        Way too much starting to happen here, but there we are...
        """
        new_scale = value / 100
        logger.info(f"Changing frame_emitter scale factor to {new_scale}")
        self.controller.scale_intrinsic_stream(self.port, new_scale)
        self.controller.stream_jump_to(self.port, self.index)

    def calibrate(self):
        self.controller.calibrate_camera(self.port)

    def clear_calibration_data(self):
        self.controller.clear_calibration_data(self.port)
        self.controller.stream_jump_to(self.port, self.index)

    def toggle_distortion_changed(self, state):
        if state == 2:
            logger.info("Apply distortion model")
            self.controller.apply_distortion(self.port, True)
            self.controller.stream_jump_to(self.port, self.index)

        else:
            logger.info("Removing distortion")
            self.controller.apply_distortion(self.port, False)
            self.controller.stream_jump_to(self.port, self.index)

    def closeEvent(self, event):
        # self.cap.release()
        super().closeEvent(event)

    def rotate_cw(self):
        self.controller.rotate_camera(self.port, 1)
        self.controller.stream_jump_to(self.port, self.index)

    def rotate_ccw(self):
        self.controller.rotate_camera(self.port, -1)
        self.controller.stream_jump_to(self.port, self.index)

    def update_enable_all_inputs(self, port, enable: bool):
        # Control widget accessibilty from controller signal to all ports
        if port == self.port:
            self.play_button.setEnabled(enable)
            self.slider.setEnabled(enable)
            self.add_grid_btn.setEnabled(enable)
            self.calibrate_btn.setEnabled(enable)
            self.camera_data_display.setEnabled(enable)
            self.clear_calibration_data_btn.setEnabled(enable)
            self.toggle_distortion.setEnabled(enable)
            self.cw_rotation_btn.setEnabled(enable)
            self.ccw_rotation_btn.setEnabled(enable)
            self.autocalibrate_btn.setEnabled(enable)
            self.target_grid_count_spin.setEnabled(enable)
            self.board_threshold_spin.setEnabled(enable)
            self.scaling_spin.setEnabled(enable)

    def autocalibrate(self):
        grid_count = self.target_grid_count_spin.value()
        board_threshold = self.board_threshold_spin.value()
        self.update_enable_all_inputs(self.port, False)
        self.controller.autocalibrate(self.port, grid_count, board_threshold)


if __name__ == "__main__":
    app = QApplication(sys.argv)
    from caliscope import __root__
    from caliscope.calibration.charuco import Charuco
    from caliscope.trackers.charuco_tracker import CharucoTracker

    # Define the input file path here.
    original_workspace_dir = Path(__root__, "tests", "sessions", "prerecorded_calibration")
    # workspace_dir = Path(
    #     __root__, "tests", "sessions_copy_delete", "prerecorded_calibration"
    # )

    # copy_contents(original_workspace_dir, workspace_dir)
    workspace_dir = Path(r"C:\Users\Mac Prible\OneDrive\caliscope\prerecorded_workflow")
    controller = Controller(workspace_dir)
    charuco = Charuco(4, 5, 11, 8.5, aruco_scale=0.75, square_size_overide_cm=5.25, inverted=True)
    charuco_tracker = CharucoTracker(charuco)
    controller.charuco_tracker = charuco_tracker

    controller.load_camera_array()
    controller.load_intrinsic_stream_manager()
    window = IntrinsicCalibrationWidget(controller=controller, port=1)
    window.resize(800, 600)
    logger.info("About to show window")
    window.show()
    sys.exit(app.exec())



from pathlib import Path
from queue import Queue
from threading import Thread

import cv2
import mediapipe as mp
import numpy as np

import caliscope.logger

# cap = cv2.VideoCapture(0)
from caliscope.packets import PointPacket
from caliscope.tracker import Tracker
from caliscope.trackers.helper import apply_rotation, unrotate_points
from caliscope.trackers.wireframe_builder import get_wireframe

logger = caliscope.logger.get(__name__)

DRAW_IGNORE_LIST = [
    "nose",
    "left_eye_inner",
    "left_eye",
    "left_eye_outer",
    "right_eye_inner",
    "right_eye",
    "right_eye_outer",
    "left_ear",
    "right_ear",
    "mouth_left",
    "mouth_right",
    "left_wrist_pose",
    "right_wrist_pose",
    "left_pinky",
    "right_pinky",
    "left_index",
    "right_index",
    "left_thumb",
    "right_thumb",
]

POINT_NAMES = {
    0: "nose",
    1: "left_eye_inner",
    2: "left_eye",
    3: "left_eye_outer",
    4: "right_eye_inner",
    5: "right_eye",
    6: "right_eye_outer",
    7: "left_ear",
    8: "right_ear",
    9: "mouth_left",
    10: "mouth_right",
    11: "left_shoulder",
    12: "right_shoulder",
    13: "left_elbow",
    14: "right_elbow",
    15: "left_wrist_pose",
    16: "right_wrist_pose",
    17: "left_pinky",
    18: "right_pinky",
    19: "left_index",
    20: "right_index",
    21: "left_thumb",
    22: "right_thumb",
    23: "left_hip",
    24: "right_hip",
    25: "left_knee",
    26: "right_knee",
    27: "left_ankle",
    28: "right_ankle",
    29: "left_heel",
    30: "right_heel",
    31: "left_foot_index",
    32: "right_foot_index",
    100: "right_wrist",
    101: "right_thumb_CMC",
    102: "right_thumb_MCP",
    103: "right_thumb_IP",
    104: "right_thumb_tip",
    105: "right_index_finger_MCP",
    106: "right_index_finger_PIP",
    107: "right_index_finger_DIP",
    108: "right_index_finger_tip",
    109: "right_middle_finger_MCP",
    110: "right_middle_finger_PIP",
    111: "right_middle_finger_DIP",
    112: "right_middle_finger_tip",
    113: "right_ring_finger_MCP",
    114: "right_ring_finger_PIP",
    115: "right_ring_finger_DIP",
    116: "right_ring_finger_tip",
    117: "right_pinky_MCP",
    118: "right_pinky_PIP",
    119: "right_pinky_DIP",
    120: "right_pinky_tip",
    200: "left_wrist",
    201: "left_thumb_CMC",
    202: "left_thumb_MCP",
    203: "left_thumb_IP",
    204: "left_thumb_tip",
    205: "left_index_finger_MCP",
    206: "left_index_finger_PIP",
    207: "left_index_finger_DIP",
    208: "left_index_finger_tip",
    209: "left_middle_finger_MCP",
    210: "left_middle_finger_PIP",
    211: "left_middle_finger_DIP",
    212: "left_middle_finger_tip",
    213: "left_ring_finger_MCP",
    214: "left_ring_finger_PIP",
    215: "left_ring_finger_DIP",
    216: "left_ring_finger_tip",
    217: "left_pinky_MCP",
    218: "left_pinky_PIP",
    219: "left_pinky_DIP",
    220: "left_pinky_tip",
}


# keep ids in distinct ranges to avoid clashes
POSE_OFFSET = 0
RIGHT_HAND_OFFSET = 100
LEFT_HAND_OFFSET = 200
FACE_OFFSET = 500


###
class HolisticTracker(Tracker):
    def __init__(self) -> None:
        # each port gets its own mediapipe context manager
        # use a dictionary of queues for passing
        self.in_queues = {}
        self.out_queues = {}
        self.threads = {}
        wireframe_spec_path = Path(Path(__file__).parent, "holistic_wireframe.toml")
        self.wireframe = get_wireframe(wireframe_spec_path, POINT_NAMES)

    @property
    def name(self):
        return "HOLISTIC"

    def run_frame_processor(self, port: int, rotation_count: int):
        # Create a MediaPipe pose instance
        with mp.solutions.holistic.Holistic(min_detection_confidence=0.8, min_tracking_confidence=0.8) as holistic:
            while True:
                frame = self.in_queues[port].get()
                # apply rotation as needed
                frame = apply_rotation(frame, rotation_count)

                height, width, color = frame.shape
                # Convert the image to RGB format
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                results = holistic.process(frame)

                # initialize variables so none will be created if no points detected
                point_ids = []
                landmark_xy = []

                if results.pose_landmarks:
                    for landmark_id, landmark in enumerate(results.pose_landmarks.landmark):
                        # mediapipe expresses in terms of percent of frame, so must map to pixel position
                        x, y = int(landmark.x * width), int(landmark.y * height)
                        if landmark.x < 0 or landmark.x > 1 or landmark.y < 0 or landmark.y > 1:
                            # ignore
                            pass
                        else:
                            point_ids.append(landmark_id + POSE_OFFSET)
                            landmark_xy.append((x, y))

                if results.right_hand_landmarks:
                    for landmark_id, landmark in enumerate(results.right_hand_landmarks.landmark):
                        # mediapipe expresses in terms of percent of frame, so must map to pixel position
                        x, y = int(landmark.x * width), int(landmark.y * height)
                        if landmark.x < 0 or landmark.x > 1 or landmark.y < 0 or landmark.y > 1:
                            # ignore
                            pass
                        else:
                            point_ids.append(landmark_id + RIGHT_HAND_OFFSET)
                            landmark_xy.append((x, y))

                if results.left_hand_landmarks:
                    for landmark_id, landmark in enumerate(results.left_hand_landmarks.landmark):
                        # mediapipe expresses in terms of percent of frame, so must map to pixel positionND_OFFSET
                        x, y = int(landmark.x * width), int(landmark.y * height)
                        if landmark.x < 0 or landmark.x > 1 or landmark.y < 0 or landmark.y > 1:
                            # ignore
                            pass
                        else:
                            point_ids.append(landmark_id + LEFT_HAND_OFFSET)
                            landmark_xy.append((x, y))

                if results.face_landmarks:
                    for landmark_id, landmark in enumerate(results.face_landmarks.landmark):
                        # mediapipe expresses in terms of percent of frame, so must map to pixel positionFSET
                        x, y = int(landmark.x * width), int(landmark.y * height)
                        if landmark.x < 0 or landmark.x > 1 or landmark.y < 0 or landmark.y > 1:
                            # ignore
                            pass
                        else:
                            point_ids.append(landmark_id + FACE_OFFSET)
                            landmark_xy.append((x, y))

                point_ids = np.array(point_ids)
                landmark_xy = np.array(landmark_xy)
                landmark_xy = unrotate_points(landmark_xy, rotation_count, width, height)
                point_packet = PointPacket(point_ids, landmark_xy)

                self.out_queues[port].put(point_packet)

    def get_points(self, frame: np.ndarray, port: int, rotation_count: int) -> PointPacket:
        if port not in self.in_queues.keys():
            self.in_queues[port] = Queue(1)
            self.out_queues[port] = Queue(1)

            self.threads[port] = Thread(
                target=self.run_frame_processor,
                args=(port, rotation_count),
                daemon=True,
            )

            self.threads[port].start()

        self.in_queues[port].put(frame)
        point_packet = self.out_queues[port].get()

        return point_packet

    def get_point_name(self, point_id) -> str:
        if point_id < FACE_OFFSET:
            point_name = POINT_NAMES[point_id]
        else:
            point_name = "face_" + str(point_id - FACE_OFFSET)
        return point_name

    def scatter_draw_instructions(self, point_id: int) -> dict:
        point_name = self.get_point_name(point_id)
        if point_name in DRAW_IGNORE_LIST:
            rules = {"radius": 0, "color": (0, 0, 0), "thickness": 0}
        elif point_name.startswith("left"):
            rules = {"radius": 5, "color": (0, 0, 220), "thickness": 3}
        elif point_name.startswith("right"):
            rules = {"radius": 5, "color": (220, 0, 0), "thickness": 3}
        else:
            rules = {"radius": 1, "color": (220, 0, 220), "thickness": 1}

        return rules

    def get_connected_points(self) -> set[tuple[int, int]]:
        return super().get_connected_points()



import sys
from collections import OrderedDict

from PySide6.QtWidgets import QApplication, QTreeWidget, QTreeWidgetItem, QVBoxLayout, QWidget

import caliscope.logger
from caliscope.controller import Controller

logger = caliscope.logger.get(__name__)


class CameraDataDisplayWidget(QWidget):
    """
    This receives a dictionary that displays the characteristics of a given camera
    """

    def __init__(self, port: int, controller: Controller):
        super().__init__()
        self.port = port
        self.controller = controller
        self.tree = QTreeWidget()

        self.place_widgets()
        self.connect_widgets()

        # make sure you populate whatever is there on load
        self.controller.push_camera_data(self.port)

    def place_widgets(self):
        layout = QVBoxLayout()
        self.setLayout(layout)
        layout.addWidget(self.tree)
        self.tree.setHeaderLabels(["Parameter", "Value"])

    def connect_widgets(self):
        self.controller.new_camera_data.connect(self.update_tree)

    def update_tree(self, port, camera_display_dict):
        # logger.info(f"Updating display tree for port {port} with camera data {camera_display_dict}")
        # port, camera_display_dict = port_camera_display_dict

        if port == self.port:
            self.tree.clear()
            # Adding top-level items
            self.add_items(None, self.tree, camera_display_dict)
            # Expand all items
            self.expand_all_items(self.tree)

    def add_items(self, parent, tree, data_dict):
        if parent is None:
            parent = tree

        for key, value in data_dict.items():
            if isinstance(value, OrderedDict):
                # For nested OrderedDict
                item = QTreeWidgetItem(parent, [key, ""])
                self.add_items(item, tree, value)
            else:
                # For direct key-value pairs
                QTreeWidgetItem(parent, [key, str(value)])

    def expand_all_items(self, tree):
        for i in range(tree.topLevelItemCount()):
            item = tree.topLevelItem(i)
            self.expand_item(item)

    def expand_item(self, item):
        item.setExpanded(True)
        for i in range(item.childCount()):
            self.expand_item(item.child(i))

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self.adjust_column_widths()

    def adjust_column_widths(self):
        total_width = self.tree.width() - 2  # Subtracting 2 pixels for border
        self.tree.setColumnWidth(0, int(total_width * 0.7))  # 70% for parameter column
        self.tree.setColumnWidth(1, int(total_width * 0.3))  # 30% for value column


if __name__ == "__main__":
    from pathlib import Path

    from caliscope import __root__

    app = QApplication(sys.argv)

    camera_data = OrderedDict(
        [
            ("size", (1920, 1080)),
            ("RMSE", 0.2),
            ("rotation_count", 3),
            (
                "intrinsic_parameters",
                OrderedDict(
                    [
                        ("focal_length_x", 1000),
                        ("focal_length_y", 1000),
                        ("optical_center_x", 500),
                        ("optical_center_y", 300),
                    ]
                ),
            ),
            (
                "distortion_coefficients",
                OrderedDict(
                    [
                        ("radial_k1", 0.01),
                        ("radial_k2", 0.02),
                        ("radial_k3", 0.03),
                        ("tangential_p1", 0.05),
                        ("tangential_p2", 0.09),
                    ]
                ),
            ),
        ]
    )

    test_path = Path(__root__, "tests", "sessions", "prerecorded_calibration")
    controller = Controller(test_path)
    controller.config.dict["camera_count"] = 1
    ex = CameraDataDisplayWidget(port=0, controller=controller)
    controller.new_camera_data.emit(0, camera_data)
    ex.show()
    sys.exit(app.exec())



import sys
from pathlib import Path

from PySide6.QtCore import Qt
from PySide6.QtGui import QPixmap
from PySide6.QtWidgets import (
    QApplication,
    QCheckBox,
    QComboBox,
    QDoubleSpinBox,
    QFileDialog,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSpinBox,
    QVBoxLayout,
    QWidget,
)

import caliscope.logger
from caliscope.calibration.charuco import Charuco
from caliscope.controller import Controller
from caliscope.gui.utils.spinbox_utils import setup_spinbox_sizing

logger = caliscope.logger.get(__name__)


class CharucoWidget(QWidget):
    def __init__(self, controller: Controller):
        super().__init__()

        logger.info("Charuco Wizard initializing")
        self.controller = controller
        self.params = self.controller.get_charuco_params()

        # add group to do initial configuration of the charuco board
        self.charuco_config = CharucoConfigGroup(self.controller)
        self.charuco_config.row_spin.valueChanged.connect(self.build_charuco)
        self.charuco_config.column_spin.valueChanged.connect(self.build_charuco)
        self.charuco_config.width_spin.valueChanged.connect(self.build_charuco)
        self.charuco_config.length_spin.valueChanged.connect(self.build_charuco)
        self.charuco_config.units.currentIndexChanged.connect(self.build_charuco)
        self.charuco_config.invert_checkbox.stateChanged.connect(self.build_charuco)

        # Build primary actions
        self.build_save_png_group()
        self.build_true_up_group()

        # Build display of board
        self.charuco_added = False  # track to handle redrawing of board
        self.build_charuco()
        self.charuco_added = True

        #################### ESTABLISH LARGELY VERTICAL LAYOUT ##############
        self.setLayout(QVBoxLayout())
        self.setWindowTitle("Charuco Board Builder")

        self.layout().addWidget(self.charuco_config)
        self.layout().setAlignment(self.charuco_config, Qt.AlignmentFlag.AlignHCenter)
        self.layout().addWidget(QLabel("<i>Top left corner is point (0,0,0) when setting capture volume origin</i>"))
        self.layout().addWidget(self.charuco_display, 2)
        self.layout().addSpacing(10)
        self.layout().addLayout(self.save_png_hbox)
        self.layout().addSpacing(10)
        self.layout().addLayout(self.true_up_hbox)
        self.layout().addWidget(QLabel("<i>Printed square size will set the scale of the capture volume</i>"))

    def build_save_png_group(self):
        # basic png save button
        self.png_btn = QPushButton("Save &png")
        self.png_btn.setMaximumSize(100, 30)

        def save_png():
            save_file_tuple = QFileDialog.getSaveFileName(
                self,
                "Save As",
                str(Path(self.controller.workspace, "charuco.png")),
                "PNG (*.png)",
            )
            print(save_file_tuple)
            save_file_name = str(Path(save_file_tuple[0]))
            if len(save_file_name) > 1:
                print(f"Saving board to {save_file_name}")
                self.charuco.save_image(save_file_name)

        self.png_btn.clicked.connect(save_png)

        # additional mirror image option
        self.png_mirror_btn = QPushButton("Save &mirror png")
        self.png_mirror_btn.setMaximumSize(100, 30)

        def save_mirror_png():
            save_file_tuple = QFileDialog.getSaveFileName(
                self,
                "Save As",
                str(Path(self.controller.workspace, "charuco_mirror.png")),
                "PNG (*.png)",
            )
            print(save_file_tuple)
            save_file_name = str(Path(save_file_tuple[0]))
            if len(save_file_name) > 1:
                print(f"Saving board to {save_file_name}")
                self.charuco.save_mirror_image(save_file_name)

        self.png_mirror_btn.clicked.connect(save_mirror_png)

        self.save_png_hbox = QHBoxLayout()
        self.save_png_hbox.addWidget(self.png_btn)
        self.save_png_hbox.addWidget(self.png_mirror_btn)

    def build_true_up_group(self):
        self.true_up_hbox = QHBoxLayout()
        self.true_up_hbox.addWidget(QLabel("Actual Printed Square Edge Length (cm):"))

        self.printed_edge_length = QDoubleSpinBox()
        self.printed_edge_length.setSingleStep(0.01)
        self.printed_edge_length.setMaximumWidth(100)
        # self.set_true_edge_length()
        overide = self.controller.config.dict["charuco"]["square_size_overide_cm"]
        self.printed_edge_length.setValue(overide)

        def update_charuco():
            self.charuco.square_size_overide_cm = round(self.printed_edge_length.value(), 2)

            logger.info(f"Updated Square Size Overide to {self.printed_edge_length.value}")
            self.controller.charuco = self.charuco
            self.controller.config.save_charuco(self.charuco)

        self.printed_edge_length.valueChanged.connect(update_charuco)

        self.true_up_hbox.layout().addWidget(self.printed_edge_length)

    def build_charuco(self):
        columns = self.charuco_config.column_spin.value()
        rows = self.charuco_config.row_spin.value()
        board_height = self.charuco_config.length_spin.value()
        board_width = self.charuco_config.width_spin.value()
        aruco_scale = self.params["aruco_scale"]
        units = self.charuco_config.units.currentText()
        square_edge_length = self.printed_edge_length.value()
        # a
        inverted = self.charuco_config.invert_checkbox.isChecked()
        dictionary_str = self.params["dictionary"]
        legacy_pattern = self.params["legacy_pattern"]

        self.charuco = Charuco(
            columns,
            rows,
            board_height,
            board_width,
            units=units,
            dictionary=dictionary_str,
            aruco_scale=aruco_scale,
            square_size_overide_cm=square_edge_length,
            inverted=inverted,
            legacy_pattern=legacy_pattern
        )

        if not self.charuco_added:
            self.charuco_display = QLabel()
            self.charuco_display.setAlignment(Qt.AlignmentFlag.AlignHCenter)

        # interesting problem comes up when scaling this... I want to switch between scaling the width and height
        # based on how these two things relate....
        if board_height > board_width:
            charuco_height = int(self.height() / 2)
            charuco_width = int(charuco_height * (board_width / board_height))
        else:
            charuco_width = int(self.width() / 2)
            charuco_height = int(charuco_width * (board_height / board_width))

        logger.info("Building charuco thumbnail...")
        try:
            charuco_img = self.charuco.board_pixmap(charuco_width, charuco_height)
            self.charuco_display.setPixmap(charuco_img)
            # Clear any previous error message
            self.charuco_display.setStyleSheet("")
            self.charuco_display.setToolTip("")
            self.controller.update_charuco(self.charuco)
        except Exception as e:
            logger.error(f"Failed to create charuco board: {str(e)}")
            error_msg = """Unable to create board with current dimensions.\n
                        The default dictionary may by too small (can be configured in config.toml file).
                        Alternatively, the aspect ratio may be too extreme.
                        """
            self.charuco_display.setPixmap(QPixmap())  # Clear the pixmap
            self.charuco_display.setText(error_msg)
            # Optional: Add some styling to make the error message stand out
            self.charuco_display.setStyleSheet("QLabel { color: red; }")
            self.charuco_display.setToolTip("Try adjusting the width and height to have a less extreme ratio")

            charuco_img = self.charuco.board_pixmap(charuco_width, charuco_height)
            self.charuco_display.setPixmap(charuco_img)



class CharucoConfigGroup(QWidget):
    def __init__(self, controller: Controller):
        super().__init__()
        self.controller = controller
        self.params = self.controller.config.dict["charuco"]

        self.column_spin = QSpinBox()
        setup_spinbox_sizing(self.column_spin, min_value=3,max_value=999, padding=10)
        self.column_spin.setValue(self.params["columns"])
        # self.column_spin.setMinimum(3)
        # self.column_spin.setMinimumWidth(50)
        # self.column_spin.setAlignment(Qt.AlignmentFlag.AlignCenter)


        self.row_spin = QSpinBox()
        self.row_spin.setValue(self.params["rows"])
        setup_spinbox_sizing(self.row_spin, min_value=4,max_value=999, padding=10)
        # self.row_spin.setMinimum(4)
        # self.row_spin.setMinimumWidth(50)
        # self.row_spin.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.width_spin = QDoubleSpinBox()
        self.width_spin.setValue(self.params["board_width"])
        setup_spinbox_sizing(self.width_spin,min_value=1, max_value=10000,padding=10)
        # self.width_spin.setMinimum(1)
        # self.width_spin.setMaximum(10000)
        # self.width_spin.setMinimumWidth(70)
        # self.width_spin.setAlignment(Qt.AlignmentFlag.AlignCenter)


        self.length_spin = QDoubleSpinBox()
        self.length_spin.setValue(self.params["board_height"])
        setup_spinbox_sizing(self.length_spin,min_value=1, max_value=10000,padding=10)
        # self.length_spin.setMinimumWidth(70)
        # self.length_spin.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.units = QComboBox()
        self.units.addItems(["cm", "inch"])
        self.units.setCurrentText(self.params["units"])

        self.invert_checkbox = QCheckBox("&Invert")
        self.invert_checkbox.setChecked(self.params["inverted"])

        #####################  HORIZONTAL CONFIG BOX  ########################
        self.config_options = QHBoxLayout()
        self.config_options.setAlignment(Qt.AlignmentFlag.AlignHCenter)

        self.charuco_config = QGroupBox("&Configure Charuco Board")
        self.setLayout(self.config_options)

        ### SHAPE GROUP    ################################################
        shape_grp = QGroupBox("row x col")
        shape_grp.setLayout(QHBoxLayout())
        shape_grp.layout().setAlignment(Qt.AlignmentFlag.AlignHCenter)

        # These are reversed from 'row x col', but this is how it works out
        shape_grp.layout().addWidget(self.row_spin)
        shape_grp.layout().addWidget(self.column_spin)

        self.config_options.addWidget(shape_grp)

        #################### SIZE GROUP #######################################
        size_grp = QGroupBox("Target Board Size")
        size_grp.setLayout(QHBoxLayout())
        size_grp.layout().setAlignment(Qt.AlignmentFlag.AlignHCenter)

        size_grp.layout().addWidget(self.width_spin)
        size_grp.layout().addWidget(self.length_spin)
        size_grp.layout().addWidget(self.units)

        self.config_options.addWidget(size_grp)

        ############################# INVERT ####################################
        self.config_options.addWidget(self.invert_checkbox)


if __name__ == "__main__":
    from caliscope import __root__
    from caliscope.calibration.charuco import Charuco
    from caliscope.helper import copy_contents
    app = QApplication(sys.argv)

    # Define the input file path here.
    original_workspace_dir = Path(__root__, "tests", "sessions", "prerecorded_calibration")

    workspace_dir = Path(r"C:\Users\Mac Prible\OneDrive\caliscope\prerecorded_workflow")
    copy_contents(original_workspace_dir, workspace_dir)
    controller = Controller(workspace_dir)
    charuco_page = CharucoWidget(controller)

    controller.load_intrinsic_stream_manager()
    window = CharucoWidget(controller=controller)
    window.resize(800, 600)
    logger.info("About to show window")
    window.show()
    sys.exit(app.exec())



import numpy as np
import pandas as pd

import caliscope.logger

logger = caliscope.logger.get(__name__)


def gap_fill_xy(xy_base: pd.DataFrame, max_gap_size=3) -> pd.DataFrame:
    """
    xy_base: dataframe which should contain the following columns:
        sync_index (or frame_index)
        frame_time
        port
        point_id
        img_loc_x
        img_loc_y

    Note: can contain other fields that will remain in the data but will not be processed.
    Examples:
        obj_loc_x
        obj_loc_y

    Note that this can take one or multiple ports, though the port field must be included even if only
    1 port is being examined.
    """
    # Initialize a DataFrame to store the data with filled gaps
    xy_filled = pd.DataFrame()

    # if "frame_index" in xy_base.columns:
    #     index_key = "frame_index"
    # else:
    #     index_key = "sync_index"
    # index_key = "frame_index"
    index_key = "sync_index"

    last_port = -1  # init to impossible value to kick off log event on first run

    # Loop through each combination of port and point_id
    for (port, point_id), group in xy_base.groupby(["port", "point_id"]):
        #### some overhead to allow sparse logging only on the first run through
        if last_port != port:
            logger.info(
                f"Gap filling for (x,y) data from port {port}. Filling gaps that are {max_gap_size} frames or less..."
            )
        last_port = port
        ### End Conditional Logging

        # Sort by frame_index to ensure the data is in order
        group = group.sort_values(index_key)

        # Create a new DataFrame with all frame_index values in the range
        all_frames = pd.DataFrame({index_key: np.arange(group[index_key].min(), group[index_key].max() + 1)})
        all_frames["port"] = port
        all_frames["point_id"] = point_id

        # Merge the actual data with the all_frames DataFrame
        merged = pd.merge(all_frames, group, on=["port", "point_id", index_key], how="left")

        # Calculate the size of each gap
        merged["gap_size"] = (
            merged["img_loc_x"].isnull().astype(int).groupby((merged["img_loc_x"].notnull()).cumsum()).cumsum()
        )

        # Remove the rows where the gap is larger than GAP_SIZE_TO_FILL
        merged = merged[merged["gap_size"] <= max_gap_size]

        # Interpolate the values for img_loc_x and img_loc_y, limit the interpolation to GAP_SIZE_TO_FILL
        merged["frame_time"] = merged["frame_time"].interpolate(method="linear", limit=max_gap_size).astype("float64")
        merged["img_loc_x"] = merged["img_loc_x"].interpolate(method="linear", limit=max_gap_size).astype("float64")
        merged["img_loc_y"] = merged["img_loc_y"].interpolate(method="linear", limit=max_gap_size).astype("float64")

        # Append to the overall DataFrame
        xy_filled = pd.concat([xy_filled, merged])

    logger.info("(x,y) gap filling complete")
    return xy_filled


def gap_fill_xyz(xyz_base: pd.DataFrame, max_gap_size=3) -> pd.DataFrame:
    """
    xyz_base: dataframe which should contain the following columns:
        sync_index
        point_id
        loc_x
        loc_y
        loc_z
    """
    # Initialize a DataFrame to store the data with filled gaps
    xyz_filled = pd.DataFrame()

    # Loop through each combination of port and point_id
    for point_id, group in xyz_base.groupby("point_id"):
        # Sort by frame_index to ensure the data is in order
        group = group.sort_values("sync_index")

        # Create a new DataFrame with all frame_index values in the range
        all_frames = pd.DataFrame({"sync_index": np.arange(group["sync_index"].min(), group["sync_index"].max() + 1)})
        all_frames["point_id"] = point_id

        # Merge the actual data with the all_frames DataFrame
        merged = pd.merge(all_frames, group, on=["point_id", "sync_index"], how="left")

        # Calculate the size of each gap
        merged["gap_size"] = (
            merged["x_coord"].isnull().astype(int).groupby((merged["x_coord"].notnull()).cumsum()).cumsum()
        )

        # Remove the rows where the gap is larger than GAP_SIZE_TO_FILL
        merged = merged[merged["gap_size"] <= max_gap_size]

        # Interpolate the values for img_loc_x and img_loc_y, limit the interpolation to GAP_SIZE_TO_FILL
        merged["x_coord"] = merged["x_coord"].interpolate(method="linear", limit=max_gap_size).astype("float64")
        merged["y_coord"] = merged["y_coord"].interpolate(method="linear", limit=max_gap_size).astype("float64")
        merged["z_coord"] = merged["z_coord"].interpolate(method="linear", limit=max_gap_size).astype("float64")

        # Append to the overall DataFrame
        xyz_filled = pd.concat([xyz_filled, merged])

    return xyz_filled



import numpy as np
import pyqtgraph.opengl as gl

import caliscope.logger
from caliscope.calibration.capture_volume.capture_volume import CaptureVolume
from caliscope.cameras.camera_array import CameraArray
from caliscope.gui.vizualize.camera_mesh import CameraMesh, mesh_from_camera

logger = caliscope.logger.get(__name__)


class CaptureVolumeVisualizer:
    """
    Can except either a single camera array or a capture volume that includes
    point_estimates. If a capture volume is supplied, point positions can
    be played back.
    """

    def __init__(self, capture_volume: CaptureVolume = None, camera_array: CameraArray = None):
        if camera_array is not None and capture_volume is None:
            self.camera_array = camera_array
            self.point_estimates = None
        else:
            self.capture_volume = capture_volume
            self.camera_array = capture_volume.camera_array
            self.point_estimates = self.capture_volume.point_estimates

        # constuct a scene
        self.scene = gl.GLViewWidget()
        self.scene.setCameraPosition(distance=4)  # the scene camera, not a real Camera
        self.sync_index = None

        self.refresh_scene()

    def refresh_scene(self):
        logger.info("refreshing capture volume scene")
        self.scene.clear()

        axis = gl.GLAxisItem()
        self.scene.addItem(axis)

        # build meshes for all cameras
        self.meshes = {}
        for port, cam in self.camera_array.cameras.items():
            print(port)
            print(cam)
            mesh: CameraMesh = mesh_from_camera(cam)
            self.meshes[port] = mesh
            self.scene.addItem(mesh)

        # self.scene.show()

        if self.point_estimates is not None:
            self.scatter = gl.GLScatterPlotItem(
                pos=np.array([0, 0, 0]),
                color=[1, 1, 1, 1],
                size=0.01,
                pxMode=False,
            )
            self.scene.addItem(self.scatter)

            self.sync_indices = np.unique(self.point_estimates.sync_indices)
            self.sync_indices = np.sort(self.sync_indices)

            self.min_sync_index = np.min(self.sync_indices)
            self.max_sync_index = np.max(self.sync_indices)

            if self.sync_index is not None:
                self.display_points(self.sync_index)

    def display_points(self, sync_index):
        """
        sync_index is provided from the dialog and linked to the slider
        it is initially set to the minimum viable sync index
        """
        self.sync_index = sync_index
        current_sync_index_flag = self.point_estimates.sync_indices == sync_index
        single_board_indices = np.unique(self.point_estimates.obj_indices[current_sync_index_flag])

        self.single_board_points = self.point_estimates.obj[single_board_indices]
        self.mean_board_position = np.mean(self.single_board_points, axis=0)
        logger.debug(f"Mean Board Position at sync index {sync_index}: {self.mean_board_position}")

        self.scatter.setData(pos=self.single_board_points)


# %%



import pandas as pd
from scipy.signal import butter, filtfilt

import caliscope.logger

logger = caliscope.logger.get(__name__)


# Define Butterworth filter functions
def butter_lowpass(cutoff, fs, order=2):
    # cutoff frequency must be less than half the frame rate
    if cutoff < fs / 2:
        logger.warning("Low pass filter failing due to excessively low fps / excessively high filter cutoff.")

    nyq = 0.5 * fs  # Nyquist Frequency
    normal_cutoff = cutoff / nyq
    b, a = butter(order, normal_cutoff, btype="low", analog=False)
    return b, a


def butter_lowpass_filter(data, cutoff, fps, order=2):
    b, a = butter_lowpass(cutoff, fps, order=order)

    # need to adjust for short input sequences
    padlen = min(len(data) - 1, 3 * (max(len(a), len(b)) - 1))
    y = filtfilt(b, a, data, padlen=padlen)
    return y


def _smooth(landmark_data: pd.DataFrame, order, fps, cutoff, coord_names, index_name) -> pd.DataFrame:
    shifted_index_name = f"{index_name}_shifted"

    # get continuous groups of a given point that can be filtered
    landmark_data = landmark_data.sort_values(by=["point_id", index_name])
    landmark_data[shifted_index_name] = landmark_data[index_name].shift(1)
    landmark_data["new_smooth_group"] = landmark_data[index_name] != landmark_data[shifted_index_name] + 1
    landmark_data["smooth_group_index"] = landmark_data["new_smooth_group"].cumsum()
    landmark_data = landmark_data.drop([shifted_index_name, "new_smooth_group"], axis=1)

    # Apply the filter to each piecewise group
    logger.info("Applying butterworth filter to point coordinates")
    for coord in coord_names:
        landmark_data[coord] = landmark_data.groupby(["smooth_group_index"])[coord].transform(
            butter_lowpass_filter, cutoff, fps, order
        )

    landmark_data = landmark_data.sort_values([index_name, "point_id"])

    return landmark_data


def _smooth_xy(xy: pd.DataFrame, order, fps, cutoff) -> pd.DataFrame:
    """
    Running this function takes a considerable amount of time. Dramatically more than
    the xyz version. So just let this be for now and don't bother implementing it
    in practice.

    Gap fill the xy, but don't bother with the smoothing for now.

    """
    # note that in future refactors, the xy coordinates my only have a frame_index and not a sync_index
    index_name = "sync_index"
    coord_names = ["img_loc_x", "img_loc_y"]

    return _smooth(xy, order, fps, cutoff, coord_names, index_name)


def smooth_xyz(xyz: pd.DataFrame, order, fps, cutoff) -> pd.DataFrame:
    index_name = "sync_index"
    coord_names = ["x_coord", "y_coord", "z_coord"]

    return _smooth(xyz, order, fps, cutoff, coord_names, index_name)



# Construct a single logger that will be used throughout
# Detail will be logged to a single file with INFO logged to the console

import logging
import os
import sys
from pathlib import Path

from PySide6 import QtCore

from caliscope import __log_dir__

# only one file handler accross package so all messages logged to one file

app_dir_file_handler = logging.FileHandler(Path(__log_dir__, "calibration.log"), "w+")
app_dir_file_handler.setLevel(logging.INFO)

file_log_format = " %(levelname)8s| %(name)30s| %(lineno)3d|  %(message)s"
file_formatter = logging.Formatter(file_log_format)
app_dir_file_handler.setFormatter(file_formatter)


console_handler = logging.StreamHandler()
console_handler.setLevel(logging.INFO)

console_log_format = " %(levelname)8s| %(name)30s| %(lineno)3d|  %(message)s"
console_formatter = logging.Formatter(console_log_format)
console_handler.setFormatter(console_formatter)

# log_level_overides = {"caliscope.cameras.live_stream": logging.INFO}


class QtHandler(logging.Handler):
    """
    Adapted from discussion here: https://stackoverflow.com/questions/24469662/how-to-redirect-logger-output-into-pyqt-text-widget
    This handler will allow a QDialog box to pick up the logger output which may be useful for a
    splash screen to show users that something is happening during big processing moments like:
    - loading/finding cameras

    - building / unbuilding synchronizer
    - performing stereocalibration
    """

    def __init__(self):
        logging.Handler.__init__(self)
        # qt_log_format =" %(levelname)s| %(name)s |%(message)s"
        qt_log_format = " %(name)s|%(message)s"
        qt_formatter = logging.Formatter(qt_log_format)
        self.setFormatter(qt_formatter)

    def emit(self, record):
        record = self.format(record)
        if record:
            XStream.stdout().write(f"{record} \n")


class XStream(QtCore.QObject):
    _stdout = None
    _stderr = None
    messageWritten = QtCore.Signal(str)

    def flush(self):
        pass

    def fileno(self):
        return -1

    def write(self, msg):
        if not self.signalsBlocked():
            self.messageWritten.emit(msg)

    @staticmethod
    def stdout():
        if not XStream._stdout:
            XStream._stdout = XStream()
            sys.stdout = XStream._stdout
        return XStream._stdout

    @staticmethod
    def stderr():
        if not XStream._stderr:
            XStream._stderr = XStream()
            sys.stderr = XStream._stderr
        return XStream._stderr


def get(name):  # as in __name__
    logger = logging.getLogger(name)

    logger.setLevel(logging.INFO)

    logger.addHandler(app_dir_file_handler)
    logger.addHandler(console_handler)

    # avoid stepping through XStream object if in debug
    if os.getenv("DEBUG") != "1":
        qt_handler = QtHandler()
        logger.addHandler(qt_handler)

    # qt_handler = QtHandler()
    # logger.addHandler(qt_handler)

    return logger



import statistics
from pathlib import Path

import cv2

import caliscope.logger
from caliscope.cameras.camera_array import CameraData
from caliscope.cameras.synchronizer import Synchronizer
from caliscope.packets import Tracker
from caliscope.recording.recorded_stream import RecordedStream
from caliscope.recording.video_recorder import VideoRecorder

logger = caliscope.logger.get(__name__)


class SynchronizedStreamManager:
    """
    The primary job of the SynchronizedStreamManager is to take in a directory of concurrently recorded video
    as well as a Tracker and produce the xy.csv file that is the foundation of both the extrinsic calibration
    as well as the point triangulation.

    Related to this, it also directs where the recorded data will go and builds the DictionaryFrameEmitter
    that will broadcast a dictionary of frames while they are being processed.

    Because of this it has substantial and broad responsibilities that include creation of:
    - streams
    - synchronizer
    - video recorder
    """

    def __init__(
        self,
        recording_dir: Path,
        all_camera_data: dict[CameraData],
        tracker: Tracker = None,
    ) -> None:
        self.recording_dir = recording_dir
        self.all_camera_data = all_camera_data
        self.tracker = tracker

        self.subfolder_name = "processed" if tracker is None else self.tracker.name
        self.output_dir = Path(self.recording_dir, self.subfolder_name)

        self.load_video_properties()
        # To be filled when loading stream tools
        self.load_stream_tools()

    def load_stream_tools(self):
        self.streams = {}

        for camera in self.all_camera_data.values():
            stream = RecordedStream(
                directory=self.recording_dir,
                port=camera.port,
                rotation_count=camera.rotation_count,
                tracker=self.tracker,
                break_on_last=True,
            )

            self.streams[camera.port] = stream

        logger.info(f"Creating synchronizer based off of streams: {self.streams}")
        self.synchronizer = Synchronizer(self.streams)
        self.recorder = VideoRecorder(self.synchronizer, suffix=self.subfolder_name)

    def process_streams(self, fps_target=None, include_video=True):
        """
        Output file will be created in a subfolder named `tracker.name`
        This will include mp4 files with visualized landmarks as well as the file `xy.csv`
        Default behavior is to process streams at the mean frame rate they were recorded at.
        But this can be overridden with a new fps_target
        """
        logger.info(f"beginning to create recording for files saved to {self.output_dir}")
        self.recorder.start_recording(
            self.output_dir,
            include_video=include_video,
            show_points=True,
            store_point_history=True,
        )

        if fps_target is None:
            fps_target = self.mean_fps

        logger.info(f"About to start playing video streams to be processed. Streams: {self.streams}")
        for port, stream in self.streams.items():
            if fps_target is not None:
                stream.set_fps_target(fps_target)

            stream.play_video()

    def load_video_properties(self):
        fps = []
        frame_count = []
        logger.info(f"About to load video properties for files stored in {self.recording_dir}")
        logger.info(f"Current camera data is: {self.all_camera_data}")
        for camera in self.all_camera_data.values():
            mp4_path = Path(self.recording_dir, f"port_{camera.port}.mp4")

            video_properties = read_video_properties(mp4_path)
            fps.append(video_properties["fps"])
            frame_count.append(video_properties["frame_count"])
            logger.info(f"loading the following video properties: {video_properties}")

        self.mean_fps = statistics.mean(fps)
        self.mean_frame_count = statistics.mean(frame_count)


def read_video_properties(source_path: Path) -> dict:
    # Dictionary to hold video properties
    properties = {}

    # Open the video file
    video = cv2.VideoCapture(str(source_path))
    logger.info(f"Attempting to open video file: {source_path}")

    # Check if video opened successfully
    if not video.isOpened():
        raise ValueError(f"Could not open the video file: {source_path}")

    # Extract video properties
    properties["frame_count"] = int(video.get(cv2.CAP_PROP_FRAME_COUNT))
    properties["fps"] = video.get(cv2.CAP_PROP_FPS)
    properties["width"] = int(video.get(cv2.CAP_PROP_FRAME_WIDTH))
    properties["height"] = int(video.get(cv2.CAP_PROP_FRAME_HEIGHT))
    properties["size"] = (properties["width"], properties["height"])

    # Release the video capture object
    video.release()

    return properties



# %%

import pickle
from dataclasses import dataclass
from pathlib import Path

import cv2
import numpy as np
from scipy.optimize import least_squares

import caliscope.logger
from caliscope.calibration.capture_volume.point_estimates import PointEstimates
from caliscope.calibration.capture_volume.set_origin_functions import (
    get_board_origin_transform,
)
from caliscope.calibration.charuco import Charuco
from caliscope.cameras.camera_array import CameraArray

logger = caliscope.logger.get(__name__)

CAMERA_PARAM_COUNT = 6


@dataclass
class CaptureVolume:
    camera_array: CameraArray
    point_estimates: PointEstimates
    stage: int = 0
    origin_sync_index: int = None

    def __post__init__():
        logger.info("Creating capture volume from estimated camera array and stereotriangulated points...")

    def _save(self, directory: Path, descriptor: str = None):
        if descriptor is None:
            pkl_name = "capture_volume_stage_" + str(self.stage) + ".pkl"
        else:
            pkl_name = "capture_volume_stage_" + str(self.stage) + "_" + descriptor + ".pkl"
        logger.info(f"Saving stage {str(self.stage)} capture volume to {directory}")
        with open(Path(directory, pkl_name), "wb") as file:
            pickle.dump(self, file)

    def get_vectorized_params(self):
        """
        Convert the parameters of the camera array and the point estimates into one long array.
        This is the required data format of the least squares optimization
        """
        camera_params = self.camera_array.get_extrinsic_params()
        combined = np.hstack((camera_params.ravel(), self.point_estimates.obj.ravel()))

        return combined

    @property
    def rmse(self):
        if hasattr(self, "least_sq_result"):
            rmse = rms_reproj_error(self.least_sq_result.fun, self.point_estimates.camera_indices)
        else:
            param_estimates = self.get_vectorized_params()
            xy_reproj_error = xy_reprojection_error(param_estimates, self)
            rmse = rms_reproj_error(xy_reproj_error, self.point_estimates.camera_indices)

        return rmse

    def get_rmse_summary(self):
        rmse_string = f"RMSE of Reprojection Overall: {round(self.rmse['overall'],2)}\n"
        rmse_string += "    by camera:\n"
        for key, value in self.rmse.items():
            if key == "overall":
                pass
            else:
                rmse_string += f"    {key: >9}: {round(float(value),2)}\n"

        return rmse_string

    def get_xy_reprojection_error(self):
        vectorized_params = self.get_vectorized_params()
        error = xy_reprojection_error(vectorized_params, self)

        return error

    def optimize(self):
        # Original example taken from https://scipy-cookbook.readthedocs.io/items/bundle_adjustment.html

        initial_param_estimate = self.get_vectorized_params()

        # get a snapshot of where things are at the start
        # initial_xy_error = xy_reprojection_error(initial_param_estimate, self)

        # logger.info(
        #     f"Prior to bundle adjustment (stage {str(self.stage)}), RMSE is: {self.rmse}"
        # )
        logger.info(f"Beginning bundle adjustment to calculated stage {self.stage+1}")
        self.least_sq_result = least_squares(
            xy_reprojection_error,
            initial_param_estimate,
            jac_sparsity=self.point_estimates.get_sparsity_pattern(),
            verbose=2,
            x_scale="jac",
            loss="linear",
            ftol=1e-8,
            method="trf",
            # xy_reprojection error takes the vectorized param estimates as first arg and capture volume as second
            args=(self,),
        )

        self.camera_array.update_extrinsic_params(self.least_sq_result.x)
        self.point_estimates.update_obj_xyz(self.least_sq_result.x)
        self.stage += 1

        logger.info(f"Following bundle adjustment (stage {str(self.stage)}), RMSE is: {self.rmse['overall']}")

    def get_xyz_points(self):
        """Get 3d positions arrived at by bundle adjustment"""
        n_cameras = len(self.camera_array.cameras)
        xyz = self.get_vectorized_params()[n_cameras * CAMERA_PARAM_COUNT :]
        xyz = xyz.reshape(-1, 3)

        return xyz

    def shift_origin(self, origin_shift_transform: np.ndarray):
        # update 3d point estimates
        xyz = self.point_estimates.obj
        scale = np.expand_dims(np.ones(xyz.shape[0]), 1)
        xyzh = np.hstack([xyz, scale])

        new_origin_xyzh = np.matmul(np.linalg.inv(origin_shift_transform), xyzh.T).T
        self.point_estimates.obj = new_origin_xyzh[:, 0:3]

        # update camera array
        for port, camera_data in self.camera_array.cameras.items():
            camera_data.transformation = np.matmul(camera_data.transformation, origin_shift_transform)

    def set_origin_to_board(self, sync_index, charuco: Charuco):
        """
        Find the pose of the charuco (rvec and tvec) from a given frame
        Transform stereopairs and 3d point estimates for this new origin
        """
        self.origin_sync_index = sync_index

        logger.info(f"Capture volume origin set to board position at sync index {sync_index}")

        origin_transform = get_board_origin_transform(self.camera_array, self.point_estimates, sync_index, charuco)
        self.shift_origin(origin_transform)


def xy_reprojection_error(current_param_estimates, capture_volume: CaptureVolume):
    """
    current_param_estimates:
        the current iteration of the vector that was originally initialized for the x0 input of least squares

    This function exists outside of the CaptureVolume class because the first argument must be the vector of parameters
    that is being adjusted by the least_squares optimization.

    """
    # Create one combined array primarily to make sure all calculations line up
    ## unpack the working estimates of the camera parameters (could be extr. or intr.)
    camera_params = current_param_estimates[: capture_volume.point_estimates.n_cameras * CAMERA_PARAM_COUNT].reshape(
        (capture_volume.point_estimates.n_cameras, CAMERA_PARAM_COUNT)
    )

    ## similarly unpack the 3d point location estimates
    points_3d = current_param_estimates[capture_volume.point_estimates.n_cameras * CAMERA_PARAM_COUNT :].reshape(
        (capture_volume.point_estimates.n_obj_points, 3)
    )

    ## create zero columns as placeholders for the reprojected 2d points
    rows = capture_volume.point_estimates.camera_indices.shape[0]
    blanks = np.zeros((rows, 2), dtype=np.float64)

    ## hstack all these arrays for ease of reference
    points_3d_and_2d = np.hstack(
        [
            np.array([capture_volume.point_estimates.camera_indices]).T,
            points_3d[capture_volume.point_estimates.obj_indices],
            capture_volume.point_estimates.img,
            blanks,
        ]
    )

    # iterate across cameras...while this injects a loop in the residual function
    # it should scale linearly with the number of cameras...a tradeoff for stable
    # and explicit calculations...

    for port, cam in capture_volume.camera_array.cameras.items():
        cam_points = np.where(capture_volume.point_estimates.camera_indices == port)
        object_points = points_3d_and_2d[cam_points][:, 1:4]

        # if a camera is being ignored, it will not show up on the parameter list
        # so you must use the port index to make sure you read the correct params
        port_index = capture_volume.camera_array.port_index[port]
        cam_matrix = cam.matrix
        rvec = camera_params[port_index][0:3]
        tvec = camera_params[port_index][3:6]
        distortions = cam.distortions

        # get the projection of the 2d points on the image plane; ignore the jacobian
        cam_proj_points, _jac = cv2.projectPoints(object_points.astype(np.float64), rvec, tvec, cam_matrix, distortions)

        points_3d_and_2d[cam_points, 6:8] = cam_proj_points[:, 0, :]

    points_proj = points_3d_and_2d[:, 6:8]

    xy_reprojection_error = (points_proj - capture_volume.point_estimates.img).ravel()
    # capture_volume.rmse = rms_reproj_error(xy_reprojection_error)
    # if round(perf_counter(), 2) * 10 % 1 == 0: # log less frequently
    #     logger.info(
    #         f"Optimizing... RMSE of reprojection = {capture_volume.rmse}"
    #     )

    # reshape the x,y reprojection error to a single vector
    return xy_reprojection_error


def rms_reproj_error(xy_reproj_error, camera_indices):
    """
    Returns a dictionary that shows the
    """
    rmse = {}
    xy_reproj_error = xy_reproj_error.reshape(-1, 2)
    euclidean_distance_error = np.sqrt(np.sum(xy_reproj_error**2, axis=1))
    rmse["overall"] = np.sqrt(np.mean(euclidean_distance_error**2))

    for port in np.unique(camera_indices):
        camera_errors = euclidean_distance_error[camera_indices == port]
        rmse[str(port)] = np.sqrt(np.mean(camera_errors**2))
    # for port in camera_indices
    # logger.info(f"Optimization run with {xy_reproj_error.shape[0]} image points")
    # logger.info(f"RMSE of reprojection is {rmse}")
    return rmse


# def load_capture_volume(session_path:Path):
#     config = get_config(session_directory)

#     camera_array = get_camera_array(config)
#     point_estimates = load_point_estimates(config)

#     capture_volume = CaptureVolume(camera_array, point_estimates)
#     capture_volume.stage = config["capture_volume"]["stage"]
#     # capture_volume.origin_sync_index = config["capture_volume"]["origin_sync_index"]

#     return capture_volume



import json
from pathlib import Path

import numpy as np
import pandas as pd

import caliscope.logger
from caliscope.trackers.tracker_enum import TrackerEnum

logger = caliscope.logger.get(__name__)


def calculate_distance(xyz_trajectory_data: pd.DataFrame, point1: str, point2: str):
    """
    Given a set of xyz trajectories from tracked and triangulated landmarks, calculate the
    mean distance between two named points, excluding outlier points for data cleanliness

    """
    # calculate the distance
    distances = np.sqrt(
        (xyz_trajectory_data[point1 + "_x"] - xyz_trajectory_data[point2 + "_x"]) ** 2
        + (xyz_trajectory_data[point1 + "_y"] - xyz_trajectory_data[point2 + "_y"]) ** 2
        + (xyz_trajectory_data[point1 + "_z"] - xyz_trajectory_data[point2 + "_z"]) ** 2
    )

    # calculate Q1, Q3 and IQR for outlier detection
    Q1 = distances.quantile(0.25)
    Q3 = distances.quantile(0.75)
    IQR = Q3 - Q1

    # filter out the outliers
    filtered_distances = distances[(distances >= Q1 - 1.5 * IQR) & (distances <= Q3 + 1.5 * IQR)]

    # calculate the average length
    average_length = filtered_distances.mean()
    return average_length


def generate_metarig_config(tracker_enum: TrackerEnum, xyz_csv_path: Path):
    """
    Stores metarig config json file within the tracker sub-directory within a recording folder
    """
    tracker = tracker_enum.value()

    xyz_trajectories = pd.read_csv(xyz_csv_path)
    json_path = Path(xyz_csv_path.parent, f"metarig_config_{tracker.name}.json")

    # for testing purposes, need to make sure that this file is not there before proceeding
    json_path.unlink(missing_ok=True)
    assert not json_path.exists()

    autorig_config = {}  # Dictionary that will become json

    # average distances across the bilateral measures
    for measure, points in tracker.metarig_bilateral_measures.items():
        logger.info(f"Calculating mean distance (IQR) for {measure}")

        mean_distance = 0
        for side in ["left", "right"]:
            point1 = f"{side}_{points[0]}"
            point2 = f"{side}_{points[1]}"

            distance = calculate_distance(xyz_trajectories, point1, point2)
            logger.info(f"Between {point1} and {point2} the mean distance is {distance}")
            mean_distance += distance / 2
        autorig_config[measure] = round(mean_distance, 4)

    # calculate distance across symmetrical measures
    for measure, points in tracker.metarig_symmetrical_measures.items():
        distance = calculate_distance(xyz_trajectories, points[0], points[1])
        autorig_config[measure] = round(distance, 4)

    # save output to file that will be in same folder as the associated xy_csv data
    with open(json_path, "w") as f:
        json.dump(autorig_config, f, indent=4)



import cv2
from PySide6.QtGui import QImage

import caliscope.logger

logger = caliscope.logger.get(__name__)


def resize_to_square(frame):
    height = frame.shape[0]
    width = frame.shape[1]

    padded_size = max(height, width)

    height_pad = int((padded_size - height) / 2)
    width_pad = int((padded_size - width) / 2)
    pad_color = [0, 0, 0]
    # pad_color = [100, 100, 100]

    frame = cv2.copyMakeBorder(
        frame,
        height_pad,
        height_pad,
        width_pad,
        width_pad,
        cv2.BORDER_CONSTANT,
        value=pad_color,
    )

    return frame


def apply_rotation(frame, rotation_count: int):
    if rotation_count == 0:
        pass
    elif rotation_count in [1, -3]:
        frame = cv2.rotate(frame, cv2.ROTATE_90_CLOCKWISE)
    elif rotation_count in [2, -2]:
        frame = cv2.rotate(frame, cv2.ROTATE_180)
    elif rotation_count in [-1, 3]:
        frame = cv2.rotate(frame, cv2.ROTATE_90_COUNTERCLOCKWISE)

    return frame


def cv2_to_qlabel(frame):
    Image = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
    # FlippedImage = cv2.flip(Image, 1)

    qt_frame = QImage(
        Image.data,
        Image.shape[1],
        Image.shape[0],
        QImage.Format.Format_RGB888,
    )
    return qt_frame



# %%
import pickle
from pathlib import Path

import numpy as np
import pandas as pd
import rtoml

import caliscope.logger
from caliscope.calibration.capture_volume.capture_volume import CaptureVolume, xy_reprojection_error
from caliscope.calibration.capture_volume.point_estimates import PointEstimates
from caliscope.calibration.charuco import Charuco

logger = caliscope.logger.get(__name__)


class QualityController:
    def __init__(self, capture_volume: CaptureVolume, charuco: Charuco = None):
        self.charuco = charuco
        self.capture_volume = capture_volume


    @property
    def data_2d(self) -> pd.DataFrame:
        """Unpack the Array Diagnostic data into a pandas dataframe format that can be
        plotted and summarized. This is all 2d data observations with their
        corresponding 3d point estimates (meaning the 3d point data is duplicated)
        """
        capture_volume_xy_error = xy_reprojection_error(
            self.capture_volume.get_vectorized_params(),
            self.capture_volume,
        ).reshape(-1, 2)
        # build out error as singular distanc

        xyz = self.capture_volume.get_xyz_points()

        euclidean_distance_error = np.sqrt(np.sum(capture_volume_xy_error**2, axis=1))
        row_count = euclidean_distance_error.shape[0]

        array_data_dict = {
            "stage": [self.capture_volume.stage] * row_count,
            "camera": self.capture_volume.point_estimates.camera_indices.tolist(),
            "sync_index": self.capture_volume.point_estimates.sync_indices.astype(
                int,
            ).tolist(),
            "charuco_id": self.capture_volume.point_estimates.point_id.tolist(),
            "img_x": self.capture_volume.point_estimates.img[:, 0].tolist(),
            "img_y": self.capture_volume.point_estimates.img[:, 1].tolist(),
            "reproj_error_x": capture_volume_xy_error[:, 0].tolist(),
            "reproj_error_y": capture_volume_xy_error[:, 1].tolist(),
            "reproj_error": euclidean_distance_error.tolist(),
            "reproj_error_sq": (euclidean_distance_error**2).tolist(),
            "obj_id": self.capture_volume.point_estimates.obj_indices.tolist(),
            "obj_x": xyz[self.capture_volume.point_estimates.obj_indices][
                :,
                0,
            ].tolist(),
            "obj_y": xyz[self.capture_volume.point_estimates.obj_indices][
                :,
                1,
            ].tolist(),
            "obj_z": xyz[self.capture_volume.point_estimates.obj_indices][
                :,
                2,
            ].tolist(),
        }

        summarized_data = pd.DataFrame(array_data_dict).astype(
            {"sync_index": "int32", "charuco_id": "int32", "obj_id": "int32"},
        )

        reproj_errors = summarized_data["reproj_error"].values
        sorted_indices = np.argsort(reproj_errors)
        ranks = np.empty_like(sorted_indices)
        ranks[sorted_indices] = np.linspace(0, 100, len(reproj_errors))
        summarized_data["reproj_error_percentile"] = ranks


        return summarized_data

    @property
    def corners_world_xyz(self) -> pd.DataFrame:
        """Convert the table of 2d data observations to a smaller table of only the individual 3d point
        estimates. These will have a number of duplicates so drop them.
        """
        corners_3d = (
            self.data_2d[["sync_index", "charuco_id", "obj_id", "obj_x", "obj_y", "obj_z"]]
            .astype({"sync_index": "int32", "charuco_id": "int32", "obj_id": "int32"})
            .drop_duplicates()
            .sort_values(by=["obj_id"])
            .reset_index()
        )

        return corners_3d

    @property
    def paired_obj_indices(self) -> np.ndarray:
        """Given a dataframe that contains all observed charuco corners across sync_indices,
        return a Nx2 matrix of paired object indices that will represent all possible
        joined lines between charuco corners for each sync_index
        """
        # get columns out from data frame for numpy calculations
        sync_indices = self.corners_world_xyz["sync_index"].to_numpy(dtype=np.int32)
        unique_sync_indices = np.unique(sync_indices)
        obj_id = self.corners_world_xyz["obj_id"].to_numpy(dtype=np.int32)

        # for a given sync index (i.e. one board snapshot) get all pairs of object ids
        paired_obj_indices = None
        for x in unique_sync_indices:
            sync_obj = obj_id[sync_indices == x]  # 3d objects (corners) at a specific sync_index
            all_pairs = cartesian_product(sync_obj, sync_obj)
            if paired_obj_indices is None:
                paired_obj_indices = all_pairs
            else:
                paired_obj_indices = np.vstack([paired_obj_indices, all_pairs])

        # paired_corner_indices will contain duplicates (i.e. [0,1] and [1,0]) as well as self-pairs ([0,0], [1,1])
        # this need to get filtered out
        reformatted_paired_obj_indices = np.zeros(
            paired_obj_indices.shape,
            dtype=np.int32,
        )
        reformatted_paired_obj_indices[:, 0] = np.min(
            paired_obj_indices,
            axis=1,
        )  # smaller on left
        reformatted_paired_obj_indices[:, 1] = np.max(
            paired_obj_indices,
            axis=1,
        )  # larger on right
        reformatted_paired_obj_indices = np.unique(
            reformatted_paired_obj_indices,
            axis=0,
        )
        reformatted_paired_obj_indices = reformatted_paired_obj_indices[
            reformatted_paired_obj_indices[:, 0] != reformatted_paired_obj_indices[:, 1]
        ]

        return reformatted_paired_obj_indices

    @property
    def corners_board_xyz(self) -> np.ndarray:
        corner_ids = self.corners_world_xyz["charuco_id"]
        corners_board_xyz = self.charuco.board.getChessboardCorners()[corner_ids]

        return corners_board_xyz

    @property
    def distance_error(self) -> pd.DataFrame:
        logger.info("Beginning to calculate distance error")

        # temp numpy frame for working calculations
        corners_world_xyz = self.corners_world_xyz[["obj_x", "obj_y", "obj_z"]].to_numpy()
        corners_board_xyz = self.corners_board_xyz

        # get the xyz positions for all pairs of corners
        corners_world_A = corners_world_xyz[self.paired_obj_indices[:, 0]]
        corners_world_B = corners_world_xyz[self.paired_obj_indices[:, 1]]
        corners_board_A = corners_board_xyz[self.paired_obj_indices[:, 0]]
        corners_board_B = corners_board_xyz[self.paired_obj_indices[:, 1]]

        # get the distance between them
        distance_world_A_B = np.sqrt(
            np.sum((corners_world_A - corners_world_B) ** 2, axis=1),
        )
        distance_board_A_B = np.sqrt(
            np.sum((corners_board_A - corners_board_B) ** 2, axis=1),
        )

        distance_world_A_B = np.round(distance_world_A_B, 5)
        distance_board_A_B = np.round(distance_board_A_B, 5)

        # calculate error (in mm)
        distance_error = distance_world_A_B - distance_board_A_B

        # wrap everything up in a dataframe for ease of processing
        distance_error = pd.DataFrame(distance_error, columns=["Distance_Error"])
        distance_error["Distance_Error_mm"] = distance_error["Distance_Error"] * 1000
        distance_error["Distance_Error_mm_abs"] = abs(
            distance_error["Distance_Error_mm"],
        )

        distance_error["corner_A"] = self.paired_obj_indices[:, 0]
        distance_error["corner_B"] = self.paired_obj_indices[:, 1]

        distance_error["world_distance"] = distance_world_A_B
        distance_error["board_distance"] = distance_board_A_B
        distance_error["percent_match"] = distance_world_A_B / distance_board_A_B
        distance_error["stage"] = self.capture_volume.stage

        logger.info("returning distance error")

        return distance_error

    @property
    def distance_error_summary(self):
        logger.info("returning summary of distance error statistics")

        summary = (
            self.distance_error.groupby("board_distance")
            .agg(
                {
                    "Distance_Error_mm_abs": ["mean", "std"],
                    "Distance_Error_mm": ["mean", "std"],
                }
            )
            .reset_index()
        )

        # flatten the multi-level column index
        summary.columns = ["_".join(col).strip() for col in summary.columns.values]
        # rename the "Distance_error_mm_abs" column to "Distance_Error_mm"
        summary["board_distance_"] = summary["board_distance_"] * 1000

        summary = summary.round(2)
        summary = summary.astype(str)

        summary["|Distance Error|"] = (
            summary["Distance_Error_mm_abs_mean"] + " (" + summary["Distance_Error_mm_abs_std"] + ")"
        )
        summary["Distance Error"] = summary["Distance_Error_mm_mean"] + " (" + summary["Distance_Error_mm_std"] + ")"

        summary = summary.rename(columns={"board_distance_": "Board Distance"})
        summary = summary[["Board Distance", "Distance Error", "|Distance Error|"]]
        return summary

    def get_filtered_data_2d(self, percentile_cutoff: float):
        """Provided a cutoff percentile value, returns a filtered_data_2d dataframe
        that only represents observations that have a reprojection error below
        that threshold. Additionally, it removes any singular 2d observations
        (i.e. those that have only one snapshot image and therefore cannot
        be localized in 3d)

        percentile_cutoff: a fraction between 0 and 1
        """
        # filter data based on reprojection error
        filtered_data_2d = self.data_2d.query(
            f"reproj_error_percentile <{percentile_cutoff*100!s}",
        )

        # get the count of obj_ids to understand how many times
        # each 3d object is represented in the 2d data
        obj_id_counts = (
            filtered_data_2d.filter(["obj_id", "camera"])
            .groupby("obj_id")
            .count()
            .rename(columns={"camera": "obj_id_count"})
        )

        # merge back into the filtered data
        filtered_data_2d = filtered_data_2d.merge(obj_id_counts, "right", on=["obj_id"])

        # remove any points that now only have 1 2d image associated with them
        filtered_data_2d = filtered_data_2d.query("obj_id_count > 1").rename(
            columns={"obj_id": "original_obj_id"},
        )
        return filtered_data_2d

    def filter_point_estimates(self, fraction_to_remove: float):
        percentile_cutoff = 1 - fraction_to_remove
        filtered_data_2d = self.get_filtered_data_2d(percentile_cutoff)

        objects_3d = (
            filtered_data_2d.filter(["original_obj_id", "obj_x", "obj_y", "obj_z"])
            .drop_duplicates()
            .reset_index()
            .drop("index", axis=1)
            .reset_index()
            .rename(columns={"index": "filtered_obj_id"})
        )

        old_new_mapping = objects_3d.filter(["filtered_obj_id", "original_obj_id"])

        filtered_data_2d = filtered_data_2d.merge(old_new_mapping, how="right", on=["original_obj_id"])

        # get revised point_estimates
        sync_indices = filtered_data_2d["sync_index"].to_numpy()
        camera_indices = filtered_data_2d["camera"].to_numpy()
        point_id = filtered_data_2d["charuco_id"].to_numpy()
        img = filtered_data_2d.filter(["img_x", "img_y"]).to_numpy()
        obj_indices = filtered_data_2d["filtered_obj_id"].to_numpy()
        obj = objects_3d.filter(["obj_x", "obj_y", "obj_z"]).to_numpy()

        filtered_point_estimates = PointEstimates(
            sync_indices=sync_indices,
            camera_indices=camera_indices,
            point_id=point_id,
            img=img,
            obj_indices=obj_indices,
            obj=obj,
        )

        self.capture_volume.point_estimates = filtered_point_estimates

        # Clear previous optimization results since they no longer correspond to the filtered point set.
        # This ensures that the next call to rmse property will recalculate the error from the current points
        # rather than using cached results that would cause dimension mismatches or inconsistent calculations.
        if hasattr(self.capture_volume, "least_sq_result"):
            delattr(self.capture_volume, "least_sq_result")

def get_capture_volume(capture_volume_pkl_path: Path) -> CaptureVolume:
    logger.info(f"loading capture volume from {capture_volume_pkl_path}")
    with open(capture_volume_pkl_path, "rb") as file:
        logger.info("beginning to load file....")
        capture_volume = pickle.load(file)
        logger.info("file loaded...")
    return capture_volume


def get_charuco(config_path) -> Charuco:
    config = rtoml.load(config_path)

    ## create charuco
    charuco = Charuco(
        columns=config["charuco"]["columns"],
        rows=config["charuco"]["rows"],
        board_height=config["charuco"]["board_height"],
        board_width=config["charuco"]["rows"],
        dictionary=config["charuco"]["dictionary"],
        units=config["charuco"]["units"],
        aruco_scale=config["charuco"]["aruco_scale"],
        square_size_overide_cm=config["charuco"]["square_size_overide_cm"],
        inverted=config["charuco"]["inverted"],
    )

    return charuco


def cartesian_product(*arrays):
    """
    Helper function for creating all possible pairs of points within a given sync_index
    https://stackoverflow.com/questions/11144513/cartesian-product-of-x-and-y-array-points-into-single-array-of-2d-points
    """
    la = len(arrays)
    dtype = np.result_type(*arrays)
    arr = np.empty([len(a) for a in arrays] + [la], dtype=dtype)
    for i, a in enumerate(np.ix_(*arrays)):
        arr[..., i] = a
    return arr.reshape(-1, la)


if __name__ == "__main__":
    # if True:
    from caliscope import __root__
    from caliscope.session.session import LiveSession

    session_directory = Path(__root__, "tests", "217")
    # config_path = Path(session_directory, "config.toml")
    # capture_volume_name = "capture_volume_stage_0.pkl"

    # get the inputs for quality control (CaptureVolume and Charuco)
    # capture_volume = get_capture_volume(Path(session_directory,capture_volume_name))
    # charuco = get_charuco(config_path)

    # create QualityControl
    session = LiveSession(session_directory)
    session.load_estimated_capture_volume()
    quality_controller = QualityController(session.capture_volume)

    logger.info(quality_controller.capture_volume.stage)

    for _ in range(5):
        logger.info("Filtering out worst fitting point estimates")
        quality_controller.filter_point_estimates(0.95)
        quality_controller.capture_volume.optimize()



# The purpose of this module is to create a class that will operate as an
# interface for a given camera. It will  store data related to
# a variety of camera specific variables and allow setting of these variables:
#
# - port
#   - an integer for a live camera
#   - note this is focused on live feeds...pre-recorded video doesn't need this
# - cv2.VideoCapture object based on port
# - Default Resolution
# - list of possible resolutions
# - exposure
# - intrinsic camera properties (to be set following calibration/on load)
#   - camera matrix
#   - distortion parameters
#
# New camera configurations
# %%

import os
import platform
import time

import cv2

import caliscope.logger

logger = caliscope.logger.get(__name__)

TEST_FRAME_COUNT = 10
MIN_RESOLUTION_CHECK = 200
MAX_RESOLUTION_CHECK = 10000


class Camera:
    # https://docs.opencv.org/3.4/d4/d15/group__videoio__flags__base.html
    # see above for constants used to access properties
    def __init__(self, port, verified_resolutions=None, connect_API=None):
        if connect_API is not None:
            self.connect_API = connect_API
        else:
            if os.name == "nt":  # windows
                self.connect_API = cv2.CAP_DSHOW
            else:  # UNIX variant
                self.connect_API = cv2.CAP_ANY

        # check if source has a data feed before proceeding...if not it is
        # either in use or fake
        logger.info(f"Attempting to connect video capture at port {port}")
        test_capture = cv2.VideoCapture(port, self.connect_API)
        for _ in range(0, TEST_FRAME_COUNT):
            good_read, frame = test_capture.read()

            # pass # dealing with this in the else statemetn below...not a real camera
        if good_read:
            logger.info(f"Good read at port {port}...proceeding")
            self.port = port
            self.capture = test_capture
            self.active_port = True

            # limit buffer size so that you are always reading the latest frame
            # from https://stackoverflow.com/questions/58293187/opencv-real-time-streaming-video-capture-is-slow-how-to-drop-frames-or-getanother thread signaled a change to mediapipe overley-sync  # noqa: E501
            self.capture.set(cv2.CAP_PROP_BUFFERSIZE, 1)

            self.ignore = False  # flag camera during single camera setup to be ignored in the future

            # sets orientation in the GUI, but otherwise does not affect the frame
            self.rotation_count = 0  # +1 for each 90 degree CW rotation, -1 for CCW

            self.set_exposure()
            self.set_default_resolution()
        else:
            # probably busy
            logger.info(f"Camera at port {port} appears to be busy")
            self.port = port
            self.capture = None
            self.active_port = False
            raise Exception(f"Not reading at port {port}...likely in use")

        # Test to see if camera is virtual
        spoof_resolution = (599, 599)
        self.size = spoof_resolution
        if self.size == spoof_resolution:
            self.virtual_camera = True
        else:
            self.virtual_camera = False

        if not self.virtual_camera:
            if verified_resolutions is None:
                self.set_possible_resolutions()
            else:
                self.verified_resolutions = verified_resolutions
            # camera initializes as uncalibrated
            self.error = None
            self.matrix = None
            self.distortions = None
            self.grid_count = None
            self.translation = None
            self.rotation = None
        if isinstance(self.verified_resolutions[0], int):
            # probably not real
            self.port = port
            self.capture = None
            self.active_port = False
            logger.info(f"Camera at port {port} may be virtual")
            raise Exception(f"{port}...likely not real")

    @property
    def exposure(self):
        return self._exposure

    @exposure.setter
    def exposure(self, value):
        """Note that OpenCV appears to change the exposure value, but
        this is not read back accurately through the getter, so just
        track it manually after updating"""
        if platform.system() == "Windows":
            self.capture.set(cv2.CAP_PROP_EXPOSURE, value)

        else:
            self.capture.set(cv2.CAP_PROP_IOS_DEVICE_EXPOSURE, value)

        self._exposure = value

    @property
    def _width(self):
        return int(self.capture.get(cv2.CAP_PROP_FRAME_WIDTH))

    @_width.setter
    def _width(self, value):
        self.capture.set(cv2.CAP_PROP_FRAME_WIDTH, value)

    @property
    def _height(self):
        return int(self.capture.get(cv2.CAP_PROP_FRAME_HEIGHT))

    @_height.setter
    def _height(self, value):
        self.capture.set(cv2.CAP_PROP_FRAME_HEIGHT, value)

    @property
    def size(self):
        return (self._width, self._height)

    @size.setter
    def size(self, value):
        """Currently, this is how the resolution is actually changed"""
        self._width = value[0]
        self._height = value[1]

    def set_default_resolution(self):
        """called at initilization before anything has changed"""
        self.default_resolution = self.size

    def set_exposure(self):
        """Need an initial value, though it does not appear that updates to
        exposure reliably read back"""
        self._exposure = self.capture.get(cv2.CAP_PROP_EXPOSURE)
        self.exposure = self._exposure  # port seemed to hold on to old exposure

    def get_nearest_resolution(self, test_width):
        """This strange little method just temporarly stores the current value
        of the resolution to be replaced at the end, then tries a value
        and then reads what resolution closest to it the capture offers,
        then returns the capture to its original state"""
        old_width = self._width
        self._width = test_width
        resolution = self.size
        self._width = old_width
        return resolution

    def set_possible_resolutions(self):
        self.verified_resolutions = []
        for resolution in RESOLUTIONS_TO_CHECK:
            # attempt to set the camera to the given resolution
            logger.info(f"Checking resolution of {resolution} at port {self.port}")
            self.size = resolution

            # if it sticks, then that resolution is verified
            if resolution == self.size:
                self.verified_resolutions.append(resolution)

    def rotate_CW(self):
        if self.rotation_count == 3:
            self.rotation_count = 0
        else:
            self.rotation_count = self.rotation_count + 1

    def rotate_CCW(self):
        if self.rotation_count == -3:
            self.rotation_count = 0
        else:
            self.rotation_count = self.rotation_count - 1

    def disconnect(self):
        self.capture.release()

    def connect(self):
        self.capture = cv2.VideoCapture(self.port, self.connect_API)

    def calibration_summary(self):
        # Calibration output presented in label on far right
        grid_count = "Grid Count:\t" + str(self.grid_count)
        size_text = "Resolution:\t" + str(self.size[0]) + "x" + str(self.size[1])

        # only grab if they exist
        if self.error and self.error != "NA":
            error_text = f"Error:\t{round(self.error,3)} "
            cam_matrix_text = "Camera Matrix:\n" + (
                "\n".join(["\t".join([str(round(float(cell), 1)) for cell in row]) for row in self.matrix])
            )
            distortion_text = "Distortion:\t" + ",".join([str(round(float(cell), 2)) for cell in self.distortions])

            # print(self.camera_matrix)
            summary = (
                grid_count
                + "\n\n"
                + error_text
                + "\n\n"
                + size_text
                + "\n\n"
                + cam_matrix_text
                + "\n\n"
                + distortion_text
            )
            return summary
        else:
            return "No Calibration Stored"


# common possibilities taken from https://en.wikipedia.org/wiki/List_of_common_resolutions
RESOLUTIONS_TO_CHECK = [
    # (352, 240),
    # (352, 288),
    # (352, 480),
    # (352, 576),
    # (352, 480),
    # (480, 480),
    # (480, 576),
    # (528, 480),
    # (544, 480),
    # (544, 576),
    (640, 480),
    # (704, 480),
    # (704, 576),
    # (720, 480),
    # (720, 576),
    # (720, 480),
    # (720, 576),
    (1280, 720),
    # (1280, 1080),
    # (1440, 1080),
    (1920, 1080),
    # (3840, 2160),
    # (7680, 4320),
]

######################### TEST FUNCTIONALITY OF CAMERAS ########################
if __name__ == "__main__":
    cam = Camera(4, verified_resolutions=[(640, 480), (1280, 720)])
    logger.info(f"Camera {cam.port} has possible resolutions: {cam.verified_resolutions}")

    for res in cam.verified_resolutions:
        logger.info(f"Testing Resolution {res}")
        logger.info("Disconnecting from camera")
        cam.disconnect()
        logger.info("Reconnecting to camera")
        cam.connect()
        logger.info(f"Setting camera size to {res}")
        cam.size = res
        logger.info("Resolution successfully updated...")

        while True:
            success, frame = cam.capture.read()
            cv2.imshow(f"Resolution: {res}; press 'q' to move to next resolution", frame)
            if cv2.waitKey(1) == ord("q"):
                cam.disconnect()
                cv2.destroyAllWindows()
                break

    cam.connect()

    # while not cam.capture.isOpened():
    #     time.sleep(.01)

    exposure_test_started = False

    start_time = time.perf_counter()

    while True:
        success, frame = cam.capture.read()
        elapsed_seconds = int(time.perf_counter() - start_time)
        logger.debug(f"{elapsed_seconds} seconds have elapsed since loop began")

        cv2.imshow("Exposure Test", frame)

        cam.exposure = -10 + elapsed_seconds

        if cv2.waitKey(1) == ord("q"):
            cam.disconnect()
            cv2.destroyAllWindows()
            break

        if elapsed_seconds > 10:
            cam.disconnect()
            cv2.destroyAllWindows()
            break

# %%
